/*
 */
/*
 * adns lookup TYPE DOMAIN [QUERY-OPTIONS]                    => [list RDATA]
 *    if no or dontknow, throws an exception, with errorCode one of
 *         ADNS permfail 300 nxdomain {No such domain}
 *         ADNS permfail 301 nodata {No such data}
 *         ADNS tempfail ERROR-CODE ERROR-NAME ERROR-STRING
 *    where
 *         ERROR-CODE is the numerical adns status value
 *         ERROR-NAME is the symbolic adns status value (in lowercase)
 *         ERROR-STRING is the result of adns_strstatus
 *
 * adns synch TYPE DOMAIN [QUERY-OPTIONS]                     => RESULTS
 *        RESULTS is [list ok|permfail|tempfail
 *                         ERROR-CODE ERROR-NAME ERROR-STRING  \
 *                         OWNER CNAME                         \
 *                         [list RDATA ...]]
 *        OWNER is the RR owner
 *        CNAME is the empty string or the canonical name if we went
 *                  via a CNAME
 *
 * adns asynch ON-YES ON-NO ON-DONTKNOW XARGS \
 *             TYPE DOMAIN \
 *             [QUERY-OPTIONS]                               => QUERY-ID
 *        calls, later,
 *           [concat ON-YES|ON-NO|ON-DONTKNOW XARGS RESULTS]
 * adns asynch-cancel QUERY-ID
 *
 * QUERY-OPTIONS are zero or more of
 *         -resolver RESOLVER  (see adns new-resolver)
 *                 default is to use a default resolver
 *         -search
 *         -usevc
 *         -quoteok-query
 *         -quoteok-anshost
 *         -quotefail-cname
 *         -cname-loose
 *         -cname-forbid
 *
 * adns new-resolver [RES-OPTIONS...]                         => RESOLVER
 *        options:
 *         -errfile stdout|stderr       (stderr is the default)
 *         -noerrprint
 *         -errcallback CALLBACK    results in  eval CALLBACK [list MESSAGE]
 *         -noenv|-debug|-logpid
 *         -checkc-entex
 *         -checkc-freq
 *         -reverse
 *         -reverse-any ZONE-A-LIKE
 *         -config CONFIG-STRING
 *
 * adns destroy-resolver RESOLVER
 */

#include "tables.h"
#include "hbytes.h"

/*---------- important types and forward declarations ----------*/

typedef struct Query Query;
typedef struct Resolver Resolver;
typedef struct OptionInfo OptionInfo;
static void asynch_check(Resolver *res);
static void asynch_sethandlers(Resolver *res, int shutdown);
static void asynch_query_dispose(Tcl_Interp *interp, Query *query);

/*---------- common resolver/query option processing ----------*/

typedef struct {
  /* this struct type is used to hold both resolver and query options */
  /* common to resolver and query: */
  unsigned long aflags;
  unsigned long sflags;
  /* resolver: */
  FILE *errfile;
  Tcl_Obj *errcallback;
  const char *config_string;
  /* query: */
  Resolver *resolver;
  const char *reverseany;
} OptionParse;

struct OptionInfo {
  const char *name;
  int (*fn)(Tcl_Interp *ip, const OptionInfo *oi, Tcl_Obj *arg,
	    OptionParse *op);
  int takesarg;
  unsigned long flags_add, flags_remove;
};

enum {
  oisf_makedefault= 0x0001,
  oisf_reverse=     0x0002
};

static int oiufn_f(const OptionInfo *oi, unsigned long *flags) {
  *flags &= ~oi->flags_remove;
  *flags |= oi->flags_add;
  return TCL_OK;
}
static int oifn_fa(Tcl_Interp *ip, const OptionInfo *oi, Tcl_Obj *arg,
		   OptionParse *op) { return oiufn_f(oi,&op->aflags); }
static int oifn_fs(Tcl_Interp *ip, const OptionInfo *oi, Tcl_Obj *arg,
		   OptionParse *op) { return oiufn_f(oi,&op->sflags); }

static int oifn_reverse_any(Tcl_Interp *ip, const OptionInfo *oi,
			    Tcl_Obj *arg, OptionParse *op) {
  return pat_string(ip,arg,&op->reverseany);
}

#define OIFA1(t,f,r) { "-" #f, oifn_fa, 0, adns_##t##_##f, r }
#define OIFA2(t,f,g) { "-" #f "-" #g, oifn_fa, 0, adns_##t##_##f##_##g, 0 }
#define OIFS(f) { "-" #f, oifn_fs, 0, oisf_##f, 0 }
#define OICA(o) { "-" #o, oifn_##o, 1 }

static int parse_options(Tcl_Interp *ip, int objc, Tcl_Obj *const *objv,
			 const OptionInfo opttable[], OptionParse *op) {
  const OptionInfo *oi;
  const void *oi_v;
  Tcl_Obj *arg;
  int rc;

  objc--; objv++;
  for (;;) {
    if (!objc--) break;
    rc= pat_enum(ip, *objv++, &oi_v, opttable, sizeof(OptionInfo),
		 "query or resolver option");
    if (rc) return rc;
    oi= oi_v;

    if (oi->takesarg) {
      if (!objc--) {
	setstringresult(ip,"missing value for option");
	return TCL_ERROR;
      }
      arg= *objv++;
    } else {
      arg= 0;
    }
    rc= oi->fn(ip,oi,arg,op);
    if (rc) return rc;
  }
  return TCL_OK;
}

/*---------- resolver management ----------*/

struct Resolver {
  int ix; /* first! */
  Tcl_Interp *interp;
  adns_state ads;
  Tcl_TimerToken timertoken;
  int maxfd;
  fd_set handling[3];
};

static int oifn_errfile(Tcl_Interp *ip, const OptionInfo *oi,
			Tcl_Obj *arg, OptionParse *op) {
  int rc;
  const char *str;
  
  rc= pat_string(ip,arg,&str);  if (rc) return rc;
  if (!strcmp(str,"stderr")) op->errfile= stderr;
  else if (!strcmp(str,"stdout")) op->errfile= stdout;
  else return staticerr(ip,"-errfile argument must be stderr or stdout",0);

  op->aflags &= ~adns_if_noerrprint;
  op->errcallback= 0;
  return TCL_OK;
}

static int oifn_errcallback(Tcl_Interp *ip, const OptionInfo *oi,
			    Tcl_Obj *arg, OptionParse *op) {
  op->errcallback= arg;
  op->aflags &= ~adns_if_noerrprint;
  op->errfile= 0;
  return TCL_OK;
}

static int oifn_config(Tcl_Interp *ip, const OptionInfo *oi,
		       Tcl_Obj *arg, OptionParse *op) {
  return pat_string(ip,arg,&op->config_string);
}

static const OptionInfo resolver_optioninfos[]= {
  OIFA1(if,noenv, 0),
  OIFA1(if,debug, adns_if_noerrprint),
  OIFA1(if,logpid, adns_if_noerrprint),
  OIFA1(if,noerrprint, adns_if_debug),
  OIFA2(if,checkc,entex),
  OIFA2(if,checkc,freq),
  OIFS(makedefault),
  OICA(errfile),
  OICA(errcallback),
  OICA(config),
  { 0 }
};

static void adnslogfn_callback(adns_state ads, void *logfndata,
			       const char *fmt, va_list al) {
  abort(); /* fixme implement adns_logcallbackfn */
}

static void destroy_resolver(Tcl_Interp *ip, Resolver *res) {
  void *query_v;
  adns_query aqu;
  
  if (res->ads) {
    /* although adns would throw these away for us, we need to
     * destroy our own data too and only adns has a list of them */
    for (;;) {
      adns_forallqueries_begin(res->ads);
      aqu= adns_forallqueries_next(res->ads, &query_v);
      if (!aqu) break;
      asynch_query_dispose(ip, query_v);
    }
    adns_finish(res->ads);
  }
  asynch_sethandlers(res,1);
  TFREE(res);
  /* fixme what about the default resolver */
}

static void destroy_resolver_idtabcb(Tcl_Interp *ip, void *resolver_v) {
  destroy_resolver(ip,resolver_v);
}

int do_adns_destroy_resolver(ClientData cd, Tcl_Interp *ip, void *res_v) {
  destroy_resolver(ip,res_v);
  tabledataid_disposing(ip,res_v,&adnstcl_resolvers);
  return TCL_OK;
}

int do_adns_new_resolver(ClientData cd, Tcl_Interp *ip,
			 int objc, Tcl_Obj *const *objv,
			 void **result) {
  OptionParse op;
  Resolver *res=0;
  int rc, i, ec;

  op.aflags= adns_if_noautosys;
  op.sflags= 0;
  op.errfile= 0;
  op.errcallback= 0;
  op.config_string= 0;
  rc= parse_options(ip,objc,objv,resolver_optioninfos,&op);
  if (rc) goto x_rc;

  res= TALLOC(sizeof(*res)); assert(res);
  res->ix= -1;
  res->interp= ip;
  res->ads= 0;
  res->timertoken= 0;
  res->maxfd= 0;
  for (i=0; i<3; i++) FD_ZERO(&res->handling[i]);

  if (op.aflags & adns_if_noerrprint) {
    op.errfile= 0;
    op.errcallback= 0;
  }

  ec= adns_init_logfn(&res->ads, op.aflags, op.config_string,
		      op.errcallback ? adnslogfn_callback : 0,
		      op.errcallback ? (void*)op.errcallback
		      : (void*)op.errfile);
  if (ec) { rc= posixerr(ip,ec,"create adns resolver"); goto x_rc; }

  if (op.errcallback)
    Tcl_IncrRefCount(op.errcallback);

  *result= res;
  return TCL_OK;

 x_rc:
  if (res) {
    if (res->ads) adns_finish(res->ads);
    TFREE(res);
  }
  return rc;
}

const IdDataSpec adnstcl_resolvers= {
  "adns-res", "adns-resolvers-table", destroy_resolver_idtabcb
};

/*---------- query, query option and answers - common stuff ----------*/

#define RRTYPE_EXACTLY(t) { #t, adns_r_##t }
#define RRTYPE_RAW(t) { #t, adns_r_##t##_raw }
#define RRTYPE_PLUS(t) { #t "+", adns_r_##t }
#define RRTYPE_MINUS(t) { #t "-", adns_r_##t##_raw }

const AdnsTclRRTypeInfo adnstclrrtypeinfos[]= {
  RRTYPE_EXACTLY(a),
  RRTYPE_EXACTLY(cname),
  RRTYPE_EXACTLY(hinfo),
  RRTYPE_EXACTLY(addr),

  RRTYPE_RAW(ns),
  RRTYPE_RAW(mx),

  RRTYPE_EXACTLY(soa),
  RRTYPE_EXACTLY(ptr),
  RRTYPE_EXACTLY(rp),

  RRTYPE_MINUS(soa),
  RRTYPE_MINUS(ptr),
  RRTYPE_MINUS(rp),
  { 0 }
};

static int oifn_resolver(Tcl_Interp *ip, const OptionInfo *oi,
			 Tcl_Obj *arg, OptionParse *op) {
  void *val_v;
  int rc;
  
  rc= pat_iddata(ip,arg,&val_v,&adnstcl_resolvers);
  if (rc) return rc;
  op->resolver= val_v;
  return TCL_OK;
}

static const OptionInfo query_optioninfos[]= {
  OIFA1(qf,search,0),
  OIFA1(qf,usevc,0),
  OIFA2(qf,quoteok,query),
  OIFA2(qf,quoteok,anshost),
  OIFA2(qf,quotefail,cname),
  OIFA2(qf,cname,loose),
  OIFA2(qf,cname,forbid),
  OICA(resolver),
  OIFS(reverse),
  { "-reverse-any", oifn_reverse_any, 1 },
  { 0 }
};

static int query_submit(Tcl_Interp *ip,
			const AdnsTclRRTypeInfo *type, const char *domain,
			int queryopts_objc, Tcl_Obj *const *queryopts_objv,
			adns_query *aqu_r, void *context, OptionParse *op) {
  struct sockaddr sa;
  static const int aftry[]= { AF_INET, AF_INET6 };
  int rc, r, ec;
  adns_state ads;
  
  op->aflags= adns_qf_owner;
  op->sflags= 0;
  op->resolver= 0; /* fixme default */
  op->reverseany= 0;
  rc= parse_options(ip, queryopts_objc,queryopts_objv, query_optioninfos,op);
  if (rc) return rc;

  if (op->reverseany || (op->sflags & oisf_reverse)) {
    const int *af;
    for (af=aftry; af < af + sizeof(af)/sizeof(*af); af++) {
      memset(&sa,0,sizeof(sa));
      sa.sa_family= *af;
      r= inet_pton(*af,domain,&sa);
      if (!r) goto af_found;
    }
    return staticerr(ip,"invalid address for adns reverse submit","");
  af_found:;
  }

  ads= op->resolver->ads;

  if (op->reverseany) {
    ec= adns_submit_reverse_any(ads, &sa, op->reverseany,
				type->number, op->aflags, context, aqu_r);
  } else if (op->sflags & oisf_reverse) {
    ec= adns_submit_reverse(ads, &sa,
			    type->number, op->aflags, context, aqu_r);
  } else {
    ec= adns_submit(ads, domain,
		    type->number, op->aflags, context, aqu_r);
  }
  if (ec)
    return posixerr(ip,ec,"submit adns query");

  return TCL_OK;
}

#define RESULTSTATUS_LLEN 4
#define RESULTLIST_LLEN 7

static void make_resultstatus(Tcl_Interp *ip, adns_status status,
			      Tcl_Obj *results[RESULTSTATUS_LLEN]) {
  results[0]= ret_string(ip, adns_errtypeabbrev(status));
  results[1]= ret_int(ip, status);
  results[2]= ret_string(ip, adns_errabbrev(status));
  results[3]= ret_string(ip, adns_strerror(status));
}

static Tcl_Obj *make_resultrdata(Tcl_Interp *ip, adns_answer *answer) {
  Tcl_Obj **rdata, *rl;
  int i, rrsz;
  adns_status st;
  char *datap, *rdatastring;
  
  rdata= TALLOC(sizeof(*rdata) * answer->nrrs);
  for (i=0, datap=answer->rrs.untyped;
       i<answer->nrrs;
       i++, datap += rrsz) {
    st= adns_rr_info(answer->type, 0,0, &rrsz, datap, &rdatastring);
    assert(!st);
    rdata[i]= ret_string(ip, rdatastring);
    free(rdatastring);
  }
  rl= Tcl_NewListObj(answer->nrrs, rdata);
  TFREE(rdata);
  return rl;
}

static void make_resultlist(Tcl_Interp *ip, adns_answer *answer,
			    Tcl_Obj *results[RESULTLIST_LLEN]) {

  make_resultstatus(ip, answer->status, results);
  assert(RESULTSTATUS_LLEN==4);
  results[4]= ret_string(ip, answer->owner);
  results[5]= ret_string(ip, answer->cname ? answer->cname : "");
  results[6]= make_resultrdata(ip, answer);
}

/*---------- synchronous query handling ----------*/

static int synch(Tcl_Interp *ip, const AdnsTclRRTypeInfo *rrtype,
		 const char *domain,
		 int objc, Tcl_Obj *const *objv, adns_answer **answer_r) {
  adns_query aqu;
  OptionParse op;
  Resolver *res;
  int rc, ec;
  
  rc= query_submit(ip,rrtype,domain,objc,objv,&aqu,0,&op);
  if (rc) return rc;

  res= op.resolver;
  ec= adns_wait(res->ads,&aqu,answer_r,0);
  assert(!ec);

  asynch_check(res);
  return TCL_OK;
}

int do_adns_lookup(ClientData cd, Tcl_Interp *ip,
		   const AdnsTclRRTypeInfo *rrtype,
		   const char *domain,
		   int objc, Tcl_Obj *const *objv,
		   Tcl_Obj **result) {
  int rc;
  adns_answer *answer;
  
  rc= synch(ip,rrtype,domain,objc,objv,&answer);  if (rc) return rc;

  if (answer->status) {
    Tcl_Obj *problem[RESULTSTATUS_LLEN];
    make_resultstatus(ip, answer->status, problem);
    *result= Tcl_NewListObj(RESULTSTATUS_LLEN, problem);
  } else {
    *result= make_resultrdata(ip, answer);
  }
  free(answer);
  return TCL_OK;
}

int do_adns_synch(ClientData cd, Tcl_Interp *ip,
		  const AdnsTclRRTypeInfo *rrtype,
		  const char *domain,
		  int objc, Tcl_Obj *const *objv,
		  Tcl_Obj **result) {
  int rc;
  adns_answer *answer;
  Tcl_Obj *results[RESULTLIST_LLEN];

  rc= synch(ip,rrtype,domain,objc,objv,&answer);  if (rc) return rc;
  make_resultlist(ip,answer,results);
  free(answer);
  *result= Tcl_NewListObj(RESULTLIST_LLEN, results);
  return TCL_OK;
}

/*---------- asynchronous query handling ----------*/

struct Query {
  int ix; /* first! */
  Resolver *res;
  adns_query aqu;
  ScriptToInvoke on_yes, on_no, on_fail;
  Tcl_Obj *xargs;
};

static void asynch_timerhandler(void *res_v) {
  Resolver *res= res_v;
  res->timertoken= 0;
  adns_processtimeouts(res->ads,0);
  asynch_check(res);
}

static void asynch_filehandler(void *res_v, int mask) {
  Resolver *res= res_v;
  int ec;
  
  ec= adns_processany(res->ads);
  if (ec) adns_globalsystemfailure(res->ads);
  asynch_check(res);
}

static void asynch_sethandlers(Resolver *res, int shutdown) {
  fd_set want[3];
  int maxfd;
  struct timeval tv_buf, *timeout;
  int i, fd;

  timeout= 0;
  maxfd= 0;
  for (i=0; i<3; i++) FD_ZERO(&want[i]);

  if (!shutdown)
    adns_beforeselect(res->ads,&maxfd,&want[0],&want[1],&want[2],
		      &timeout,&tv_buf,0);

  for (fd= 0; fd < maxfd || fd < res->maxfd; fd++)
    for (i=0; i<3; i++)
      if (!!FD_ISSET(fd, &res->handling[i])
	  != !!FD_ISSET(fd, &want[i])) {
	int mask=0;
	if (FD_ISSET(fd, &want[0])) mask |= TCL_READABLE;
	if (FD_ISSET(fd, &want[1])) mask |= TCL_WRITABLE;
	if (FD_ISSET(fd, &want[2])) mask |= TCL_EXCEPTION;
	if (mask) Tcl_CreateFileHandler(fd,mask,asynch_filehandler,res);
	else Tcl_DeleteFileHandler(fd);
      }

  Tcl_DeleteTimerHandler(res->timertoken);

  if (timeout) {
    int milliseconds;

    if (timeout->tv_sec >= INT_MAX/1000 - 1)
      milliseconds= INT_MAX;
    else
      milliseconds= timeout->tv_sec * 1000 +
	(timeout->tv_usec + 999) / 1000;
    
    res->timertoken=
      Tcl_CreateTimerHandler(milliseconds,asynch_timerhandler,res);
  }
}

static void asynch_check(Resolver *res) {
  Tcl_Interp *interp= res->interp;
  adns_query aqu;
  adns_answer *answer;
  void *query_v;
  Query *query;
  ScriptToInvoke *si;
  int ec;
  Tcl_Obj *results[RESULTLIST_LLEN];

  for (;;) {
    ec= adns_check(res->ads, &aqu, &answer, &query_v);
    if (ec==ESRCH || ec==EAGAIN) break;
    assert(!ec);
    query= query_v;

    query->aqu= 0;
    tabledataid_disposing(interp, query, &adnstcl_queries);

    si= (!answer->status ? si= &query->on_yes
	 : answer->status > adns_s_max_tempfail ? &query->on_no
	 : &query->on_fail);

    make_resultlist(interp, answer, results);
    free(answer);
    scriptinv_invoke(si, RESULTLIST_LLEN, results);
    asynch_query_dispose(interp, query);
  }

  asynch_sethandlers(res,0);
}
    
int do_adns_asynch(ClientData cd, Tcl_Interp *ip,
		   Tcl_Obj *on_yes, Tcl_Obj *on_no,
		   Tcl_Obj *on_fail, Tcl_Obj *xargs,
		   const AdnsTclRRTypeInfo *rrtype, const char *domain,
		   int objc, Tcl_Obj *const *objv, void **result) {
  Query *query;
  int rc;
  Resolver *res=0;
  OptionParse op;
  
  query= TALLOC(sizeof(*query));
  query->ix= -1;
  query->aqu= 0;
  scriptinv_init(&query->on_yes);
  scriptinv_init(&query->on_no);
  scriptinv_init(&query->on_fail);
  query->xargs= 0;

  rc= query_submit(ip,rrtype,domain,objc,objv,&query->aqu,query,&op);
  if (rc) goto x_rc;

  res= op.resolver;

  rc= scriptinv_set(&query->on_yes, ip,on_yes);   if (rc) goto x_rc;
  rc= scriptinv_set(&query->on_no,  ip,on_no);    if (rc) goto x_rc;
  rc= scriptinv_set(&query->on_fail,ip,on_fail);  if (rc) goto x_rc;
  query->xargs= xargs;
  Tcl_IncrRefCount(xargs);
  *result= query;

  return TCL_OK;

 x_rc:
  if (query) asynch_query_dispose(ip, query);
  if (res) asynch_sethandlers(res,0);
  return rc;
}

int do_adns_asynch_cancel(ClientData cd, Tcl_Interp *ip, void *query_v) {
  Query *query= query_v;
  Resolver *res= query->res;
  asynch_query_dispose(ip, query);
  asynch_sethandlers(res,0);
  return TCL_OK;
}

static void asynch_query_dispose(Tcl_Interp *interp, Query *query) {
  tabledataid_disposing(interp, query, &adnstcl_queries);
  scriptinv_cancel(&query->on_yes);
  scriptinv_cancel(&query->on_no);
  scriptinv_cancel(&query->on_fail);
  if (query->xargs) Tcl_DecrRefCount(query->xargs);
  if (query->aqu) adns_cancel(query->aqu);
  TFREE(query);
}

static void destroy_query_idtabcb(Tcl_Interp *interp, void *query_v) {
  asynch_query_dispose(interp, query_v);
}

const IdDataSpec adnstcl_queries= {
  "adns", "adns-query-table", destroy_query_idtabcb
};
